-[ c library "layout.c", header "layout.h" ]

-[ lheader "parser.h" ]

-[public] typedef struct layout_metrics layout_metrics_t;

-[public] struct layout_metrics {
	int page_width;
	void (*font_metrics)(layout_metrics_t *,int, wchar_t,int *,int *,int*);
	int (*bullet_indent)(layout_metrics_t *,int);
	int (*text_style)(layout_metrics_t *,int);
	int (*header_style)(layout_metrics_t *,int);
	wchar_t bullet_chr;
	int paragraph_spacing;
	int link_style;
	int preformatted_style;
	int (*tabstop)(layout_metrics_t*, int x);
	int line_spacing;
};


-[public] int layout_doc(char *txt,layout_metrics_t *metrics,text_part_t **result) {
	text_part_t *parts;
	int len;
	layout_t L;
	fill_vtable(&L);
	L.metrics= metrics;

	L.outbuf.prev= L.outbuf.next= &L.outbuf;
	L.outbuf.nparts= 0;

	L.nchars= 0;
	L.schars= 200;
	-[ L.chars= allocate array L.schars ]
	L.x= 0;
	L.y= 0;

	L.ascent= L.descent= 0;

	L.left= 0;
	L.right= metrics->page_width;
	L.last_link_id= 1;

	parse_doc(txt, &L.vtable);
	finish_row(&L);

	parts= collect_buffer_destroy(&L,&len);
	free(L.chars);
	*result= parts;
	return len;
}

-[private] void fill_vtable(layout_t *L) {
	parser_sink_t *s= &L->vtable;
	s->bullet= (void*) bullet;
	s->pbreak= (void*) pbreak;
	s->link= (void*) link;
	s->text= (void*) text;
	s->formatted= (void*) formatted;
	s->header= (void*) header;
	s->tag= (void*) ignore_tag;
}

-[private] void ignore_tag(layout_t *L, char* tag) { }


-[private] void bullet(layout_t *L, int level) {
	unsigned g;
	finish_row(L);
	L->left= L->metrics->bullet_indent(L->metrics, level);
	g= find_glyph(L, L->metrics->bullet_chr, 0);
	add_char(L, g, 0, L->left - glyphs[g].width );
	L->x= L->left;
}

-[private] void pbreak(layout_t *L) {
	finish_row(L);
	L->left= 0;
	L->x= 0;
	L->y+= L->metrics->paragraph_spacing;
}

-[private] void link(layout_t *L,int kind,char *txt,int len) {
	int style,aux;
	if (kind!='L')
		return ;
	/* FIXME: maybe use different styles for links
	   bound to Xs and Ws? */
	style= L->metrics->link_style;
	aux= L->last_link_id++;
	layout_text(L, style, txt,len, aux);
}


/* style here is translated style. i.e. device style */
-[private] void layout_text
	( layout_t *L,int style,char *txt, int len,int aux)
{
	wchar_t *wstr;
	int i;
	wstr= utf8buf2wstr(txt,len);
	for(i=0;wstr[i];i++) {
		wchar_t chr;
		unsigned g;
		chr= wstr[i];
		if (!iswprint(chr))
			chr=' ';
		if (chr==' ' && L->x==L->left)
			continue;
		g= find_glyph(L,chr, style);
		if (L->x+glyphs[g].width>L->right)  {
			int j;
			int npurged;
			int first_purged;

			if (!L->nchars || L->x==L->left)
				goto add_anyway;
	
			if (iswspace(chr)) {
				/* trying to append a space at the end of a line
				and it doesn`t fit. so we ignore it :) */
				finish_row(L);
				continue;
			}

			if (iswspace(glyphs[L->chars[L->nchars-1].glyph].chr)) {
				/* the previous char was a space so we finish
				the row at where it is and then add the char */
				finish_row(L);
				goto add_anyway;
			}

			/* ok. at this point we know that the previous char was not
			a space and this one is not either. so, we got some back-
			tracking to do. */
			for(j=L->nchars-1;j>0;j--) 
				if (iswspace(glyphs[L->chars[j].glyph].chr))
					break;

			if (j==0) { 
				/* we found ourselves a great big word that doesnt
				fit anywhere, so we will char-wrap it instead. */
				finish_row(L);
				goto add_anyway;
			}
			first_purged= j+1;
			npurged= L->nchars-first_purged;

			/* ok. if this string is not long enough. i.e. different text()
			calls resulted in one continuous word to form, then backing
			up on the wstr array wont do any good because we can not
			get back to the original characters. so, I have to store the
			stuff in an auxilliary buffer and then re-transfer them to the
			row array. instead of using an auxilliary buffer, I will re-use
			the row array because I know that finish_row wont deallocate
			or overwrite those positions and I am a jerk. */
			L->nchars -= npurged;
			recalculate_row_dimensions(L);
			finish_row(L);
			
			for(j=0;j<npurged;j++) {
				xchar_t *xc= L->chars + j + first_purged;
				add_char(L, xc->glyph, xc->aux, L->x);
				L->x+= glyphs[xc->glyph].width;
			}
			/* now I silently fall thru to add the current character.
			    you cool with that? */
		}
	add_anyway:
		add_char( L, g, aux, L->x );
		L->x += glyphs[g].width;
	}
	free(wstr);
}

-[private] void recalculate_row_dimensions(layout_t *L) {
	int i;
	L->ascent= L->descent= 0;
	for(i=0;i<L->nchars;i++) {
		max2first(&L->ascent, glyphs[L->chars[i].glyph].fascent);
		max2first(&L->descent, glyphs[L->chars[i].glyph].fdescent);
	}
}

-[private] void text(layout_t *L,int style,char *txt, int len) {
	style= L->metrics->text_style(L->metrics,style);
	layout_text(L,style, txt,len, 0);
}

-[private] void formatted(layout_t *L,char *txt, int len) {
	wchar_t *wstr;
	int i;
	int style;

	finish_row(L);
	L->left= 0;
	L->x= 0;
	style= L->metrics->preformatted_style;

	wstr= utf8buf2wstr(txt,len);
	for(i=0;wstr[i];i++) {
		wchar_t chr;
		int width;
		unsigned g;
		chr= wstr[i];
		if (chr=='\t') {
			int ts;
			g= find_glyph(L, ' ', style);
			ts= L->metrics->tabstop(L->metrics,L->x);
			width= ts-L->x;
		} else {
			if (chr=='\n')
				continue;
			g= find_glyph(L,chr,style);
			width= glyphs[g].width;
		}
		add_char(L, g, 0, L->x);
		L->x+= width;
	}
	free(wstr);
	finish_row(L);
}

-[private] void header(layout_t *L,int level,char *txt, int len) {
	int style;
	finish_row(L);
	style= L->metrics->header_style(L->metrics, level);
	layout_text(L,style, txt,len,0);
	finish_row(L);
}

-[private] typedef struct {
	unsigned int glyph;
	unsigned short  x;
	unsigned int aux;
} xchar_t;

-[public] typedef struct {
	int ascent;
	int style;
	int aux;
	char *text;
} disp_text_t;
	
-[public] typedef struct {
	int opcode;
	int width,height;
	int x,y;
	union {
		disp_text_t text;
	} d;
} text_part_t;

-[private] typedef struct outbuf {
	struct outbuf *prev,*next;
	int nparts;
	text_part_t parts[40];
} outbuf_t;

-[private] typedef struct {
	parser_sink_t vtable;
	layout_metrics_t *metrics;

	int nchars,schars;
	xchar_t *chars;

	int right;
	int left;
	int ascent, descent;
	int x,y;

	outbuf_t outbuf;

	int last_link_id;
} layout_t;



-[private] void add_char(layout_t *L, unsigned g, int aux, int x) {
	xchar_t *xc;
	-[ make space in row ]
	xc= L->chars+L->nchars;
	L->nchars++;
	xc->glyph= g;
	xc->aux= aux;
	xc->x= x;
	max2first(&L->ascent, glyphs[g].fascent);
	max2first(&L->descent, glyphs[g].fdescent);
}

-[private] void max2first(int *A,int B) { if (B>*A) *A= B; }

-[private] void finish_row(layout_t *L) {
	int s,e;
	if (L->nchars) {
		s= 0;

	again:
		e= s+1;
		while(e<L->nchars) {
			if (glyphs[L->chars[e].glyph].style!=glyphs[L->chars[s].glyph].style)
				break;
			if (L->chars[e].x!=L->chars[e-1].x+glyphs[L->chars[e-1].glyph].width)
				break;
			if (L->chars[e].aux!=L->chars[s].aux)
				break;
			e++;
		}

		output_text_part(L,s,e);
		if (e<L->nchars) {
			s= e;
			goto again;
		}
	}

	L->nchars= 0;
	L->x= L->left;
	L->y= L->y + L->ascent + L->descent + L->metrics->line_spacing;
	L->ascent= L->descent= 0;
}

-[private] void output_text_part(layout_t *L,int s,int e) {
	text_part_t  T;
	int len;
	outbuf_t *O;
	T.opcode= 'T';
	T.y= L->y;
	T.d.text.ascent= L->ascent;
	T.d.text.aux=L->chars[s].aux;
	T.d.text.style= glyphs[L->chars[s].glyph].style;
	T.x= L->chars[s].x;
	len= text_part_chars(L->chars+s, e-s, 0);
	T.d.text.text= malloc(len);
	text_part_chars(L->chars+s, e-s, T.d.text.text);

	T.height= L->ascent + L->descent;
	T.width= L->chars[e-1].x
		+ glyphs[L->chars[e-1].glyph].width
		- L->chars[s].x;

	O= L->outbuf.prev;
	if (O->nparts==sizeof(O->parts)/sizeof(T)) {
		-[allocate object O]
		-[ clist/insert O before (&L->outbuf) ]
	}
	O->parts[O->nparts++]= T;
}


-[private] int text_part_chars(xchar_t *d,int l,char *buf) {
	int i;
	int L=0;
	for(i=0;i<l;i++) {
		char *u= (char*) glyphs[d[i].glyph].utf8;
		int xl= strlen(u);
		if (buf)
			memcpy(buf+L, u, xl);
		L+= xl;
	}
	if (buf)
		buf[L]= 0;
	return L+1;
}

=[ make space in row ] {
	if (L->nchars==L->schars) {
		L->schars *= 2;
		-[ reallocate array L->chars to size L->schars ]
	}
}


=[]

-[header "stdlib.h"]
-[lheader "utf8.h"]
-[header "wctype.h" ]
-[header "string.h" ]


-[private] text_part_t *collect_buffer_destroy(layout_t *L,int *Lr) {
	int len;
	text_part_t *P;
	len= collect_buffer_nodestroy(&L->outbuf, 0);
	-[ P = allocate array len ]
	collect_buffer_nodestroy(&L->outbuf, P);
	while(1) {
		outbuf_t *B= L->outbuf.prev;
		if (B==&L->outbuf)
			break;
		-[clist/remove B]
		free(B);
	}
	*Lr= len;
	return P;
}


-[private] int collect_buffer_nodestroy(outbuf_t *L,text_part_t *B) {
	outbuf_t *N= L;
	int s= 0;
	do {
		if (B && N->nparts) 
			memcpy(B+s, N->parts,
				N->nparts*sizeof(N->parts[0]));
		s+= N->nparts;
		N= N->next;
	} while(N!=L);
	return s;
}